@@ -16,22 +16,22 @@ import textwrap
 from itertools import chain
 import pickle
 import ipywidgets as widgets
+import re
 
 # Clustering
-from bertopic import BERTopic
-from bertopic.backend import OpenAIBackend
 from hdbscan import HDBSCAN
+import umap
 
 # Local imports
 if __package__ is None or __package__ == '':
     # uses current directory visibility
-    from llm import multi_query_gpt_wrapper, calc_cost_by_tokens
+    from llm import multi_query_gpt_wrapper, calc_cost_by_tokens, get_embeddings
     from prompts import *
     from concept import Concept
     from __init__ import MatrixWidget, ConceptSelectWidget
 else:
     # uses current package visibility
-    from .llm import multi_query_gpt_wrapper, calc_cost_by_tokens
+    from .llm import multi_query_gpt_wrapper, calc_cost_by_tokens, get_embeddings
     from .prompts import *
     from .concept import Concept
     from .__init__ import MatrixWidget, ConceptSelectWidget
@@ -73,29 +73,33 @@ def pretty_print_dict_list(d_list):
 
 def cluster_helper(in_df, doc_col, doc_id_col, min_cluster_size, cluster_id_col, embed_model_name):
     # OpenAI embeddings with HDBSCAN clustering
-    embedding_model = OpenAIBackend(embed_model_name)
+    id_vals = in_df[doc_id_col].tolist()
+    text_vals = in_df[doc_col].tolist()
 
-    hdbscan_model = HDBSCAN(
+    embeddings = get_embeddings(embed_model_name, text_vals)
+    umap_model = umap.UMAP(
+        n_neighbors=15,
+        n_components=5,
+        min_dist=0.0,
+        metric='cosine',
+    )
+    umap_embeddings = umap_model.fit_transform(embeddings)
+    hdb = HDBSCAN(
         min_cluster_size=min_cluster_size, 
         metric='euclidean', 
         cluster_selection_method='leaf', 
         prediction_data=True
     )
-    topic_model = BERTopic(
-        embedding_model=embedding_model, 
-        hdbscan_model=hdbscan_model
-    )
+    res = hdb.fit(umap_embeddings)
+    clusters = res.labels_
 
-    id_vals = in_df[doc_id_col].tolist()
-    text_vals = in_df[doc_col].tolist()
-    clusters, probs = topic_model.fit_transform(text_vals)
     rows = list(zip(id_vals, text_vals, clusters)) # id_col, text_col, cluster_id_col
     cluster_df = pd.DataFrame(rows, columns=[doc_id_col, doc_col, cluster_id_col])
     cluster_df = cluster_df.sort_values(by=[cluster_id_col])
     
     return cluster_df
 
-def save_progress(sess, df, step_name, start, res, model_name, debug=True):
+def save_progress(sess, df, step_name, start, res, model_name, debug=False):
     # Save df to session
     if (sess is not None) and (df is not None):
         k = sess.get_save_key(step_name)
@@ -107,7 +111,7 @@ def save_progress(sess, df, step_name, start, res, model_name, debug=True):
     # Gets cost
     calc_cost(res, model_name, step_name, sess, debug=debug)
 
-def get_timing(start, step_name, sess, debug=True):
+def get_timing(start, step_name, sess, debug=False):
     if start is None:
         return
     elapsed = time.time() - start
@@ -118,13 +122,13 @@ def get_timing(start, step_name, sess, debug=True):
         k = sess.get_save_key(step_name)
         sess.time[k] = elapsed
 
-def calc_cost(results, model_name, step_name, sess, debug=True):
+def calc_cost(results, model_name, step_name, sess, debug=False):
     # Calculate cost with API results and model name
     if results is None:
         return
     # Cost estimation
-    in_tokens = np.sum([res.llm_output["token_usage"]["prompt_tokens"] for res in results])
-    out_tokens = np.sum([res.llm_output["token_usage"]["completion_tokens"] for res in results])
+    in_tokens = np.sum([res.usage.prompt_tokens for res in results])
+    out_tokens = np.sum([res.usage.completion_tokens for res in results])
     in_token_cost, out_token_cost = calc_cost_by_tokens(model_name, in_tokens, out_tokens)
     total_cost = in_token_cost + out_token_cost
     if debug:
@@ -171,7 +175,11 @@ async def distill_filter(text_df, doc_col, doc_id_col, model_name, n_quotes=3, s
     
     # Run prompts
     prompt_template = filter_prompt
-    res_text, res_full = await multi_query_gpt_wrapper(prompt_template, arg_dicts, model_name)
+    if sess is not None:
+        rate_limits = sess.rate_limits
+    else:
+        rate_limits = None
+    res_text, res_full = await multi_query_gpt_wrapper(prompt_template, arg_dicts, model_name, rate_limits=rate_limits)
 
     # Process results
     ex_ids = [ex_id for ex_id in text_df[doc_id_col].tolist()]
@@ -183,7 +191,7 @@ async def distill_filter(text_df, doc_col, doc_id_col, model_name, n_quotes=3, s
             rows.append([ex_id, cur_filtered])
     quote_df = pd.DataFrame(rows, columns=[doc_id_col, doc_col])
     
-    save_progress(sess, quote_df, step_name="distill_filter", start=start, res=res_full, model_name=model_name)
+    save_progress(sess, quote_df, step_name="Distill-filter", start=start, res=res_full, model_name=model_name)
     return quote_df
 
 
@@ -225,7 +233,11 @@ async def distill_summarize(text_df, doc_col, doc_id_col, model_name, n_bullets=
     
     # Run prompts
     prompt_template = summarize_prompt
-    res_text, res_full = await multi_query_gpt_wrapper(prompt_template, arg_dicts, model_name)
+    if sess is not None:
+        rate_limits = sess.rate_limits
+    else:
+        rate_limits = None
+    res_text, res_full = await multi_query_gpt_wrapper(prompt_template, arg_dicts, model_name, rate_limits=rate_limits)
 
     # Process results
     for ex_id, res in zip(all_ex_ids, res_text):
@@ -236,7 +248,7 @@ async def distill_summarize(text_df, doc_col, doc_id_col, model_name, n_bullets=
                 rows.append([ex_id, bullet])
     bullet_df = pd.DataFrame(rows, columns=[doc_id_col, doc_col])
 
-    save_progress(sess, bullet_df, step_name="distill_summarize", start=start, res=res_full, model_name=model_name)
+    save_progress(sess, bullet_df, step_name="Distill-summarize", start=start, res=res_full, model_name=model_name)
     return bullet_df
 
 
@@ -272,7 +284,7 @@ async def cluster(text_df, doc_col, doc_id_col, cluster_id_col="cluster_id", min
         # Cluster and group by clusters
         cluster_df = cluster_helper(text_df, doc_col, doc_id_col, min_cluster_size=min_cluster_size, cluster_id_col=cluster_id_col, embed_model_name=embed_model_name)
 
-    save_progress(sess, cluster_df, step_name="cluster", start=start, res=None, model_name=None)
+    save_progress(sess, cluster_df, step_name="Cluster", start=start, res=None, model_name=None)
     return cluster_df
 
 
@@ -289,7 +301,7 @@ def dict_to_json(examples):
 # Output: 
 # - concepts: dict (concept_id -> concept dict)
 # - concept_df: DataFrame (columns: doc_id, doc, concept_id, concept_name, concept_prompt)
-async def synthesize(cluster_df, doc_col, doc_id_col, model_name, cluster_id_col="cluster_id", concept_col_prefix="concept", n_concepts=None, batch_size=None, verbose=True, pattern_phrase="unifying pattern", dedupe=True, seed=None, sess=None):
+async def synthesize(cluster_df, doc_col, doc_id_col, model_name, cluster_id_col="cluster_id", concept_col_prefix="concept", n_concepts=None, batch_size=None, verbose=False, pattern_phrase="unifying pattern", dedupe=True, seed=None, sess=None, return_logs=False):
     # Synthesis operates on "doc" column for each cluster_id
     # Concept object is created for each concept
     start = time.time()
@@ -348,17 +360,19 @@ async def synthesize(cluster_df, doc_col, doc_id_col, model_name, cluster_id_col
 
     # Run prompts
     prompt_template = synthesize_prompt
-    res_text, res_full = await multi_query_gpt_wrapper(prompt_template, arg_dicts, model_name)
+    if sess is not None:
+        rate_limits = sess.rate_limits
+    else:
+        rate_limits = None
+    res_text, res_full = await multi_query_gpt_wrapper(prompt_template, arg_dicts, model_name, rate_limits=rate_limits)
 
     # Process results
     concepts = {}
     rows = []
+    logs = ""
     for cur_cluster_id, res in zip(cluster_ids, res_text):
         cur_concepts = json_load(res, top_level_key="patterns")
         if cur_concepts is not None:
-            # concepts.extend(cur_concepts)
-            # cur_examples_dict = {ex_to_id[cur_ex]: cur_ex for cur_ex in cur_examples}
-            
             for concept_dict in cur_concepts:
                 ex_ids = concept_dict["example_ids"]
                 ex_ids = set(ex_ids) # remove duplicates
@@ -376,10 +390,13 @@ async def synthesize(cluster_df, doc_col, doc_id_col, model_name, cluster_id_col
                     if cur_key in ex_id_to_ex:
                         row = [ex_id, ex_id_to_ex[cur_key], concept.id, concept.name, concept.prompt]
                         rows.append(row)
+            # Print intermediate results
+            examples = cluster_dfs[cur_cluster_id][doc_col].tolist()
+            concepts_formatted = pretty_print_dict_list(cur_concepts)
+            cur_log = f"\n\nInput examples: {examples}\nOutput concepts: {concepts_formatted}"
+            logs += cur_log
             if verbose:
-                examples = cluster_dfs[cur_cluster_id][doc_col].tolist()
-                concepts_formatted = pretty_print_dict_list(cur_concepts)
-                print(f"\n\nInput examples: {examples}\nOutput concepts: {concepts_formatted}")
+                print(cur_log)
     # doc_id, text, concept_id, concept_name, concept_prompt
     concept_df = pd.DataFrame(rows, columns=[doc_id_col, doc_col, concept_col_prefix, f"{concept_col_prefix}_name", f"{concept_col_prefix}_prompt"])
 
@@ -387,38 +404,58 @@ async def synthesize(cluster_df, doc_col, doc_id_col, model_name, cluster_id_col
     if dedupe:
         concept_df = dedupe_concepts(concept_df, concept_col=f"{concept_col_prefix}_namePrompt")
 
-    save_progress(sess, concept_df, step_name="synthesize", start=start, res=res_full, model_name=model_name)
+    save_progress(sess, concept_df, step_name="Synthesize", start=start, res=res_full, model_name=model_name)
     # Save to session if provided
     if sess is not None:
         for c_id, c in concepts.items():
             sess.concepts[c_id] = c
+    
+    if return_logs:
+        return concept_df, logs
     return concept_df
 
 def dedupe_concepts(df, concept_col):
     # Remove duplicate concept rows
     return df.drop_duplicates(subset=[concept_col])
 
-def pretty_print_merge_results(merged):
+def get_merge_results(merged):
+    out = ""
     for m in merged:
         orig_concepts = m["original_themes"]
-        print(f"[{orig_concepts}] --> {m['merged_theme_name']}: {m['merged_theme_prompt']}")
+        cur_out = f"\t{orig_concepts} --> {m['merged_theme_name']}: {m['merged_theme_prompt']}"
+        out += cur_out + "\n"
+    return out
 
 # Input: concept_df (columns: doc_id, text, concept_id, concept_name, concept_prompt)
 # Parameters: n_concepts
 # Output: 
-# - concepts: dict (concept_id -> concept dict)
+# - concepts: dict (concept_id -> Concept)
 # - concept_df: DataFrame (columns: doc_id, text, concept_id, concept_name, concept_prompt)
-async def review(concepts, concept_df, concept_col, concept_col_prefix, model_name, debug=True, sess=None):
+async def review(concepts, concept_df, concept_col_prefix, model_name, debug=False, sess=None, return_logs=False):
     # Model is asked to review the provided set of concepts
-    concepts_out, concept_df_out, removed = await review_remove(concepts, concept_df, concept_col, concept_col_prefix, model_name=model_name, sess=sess)
-    concepts_out, concept_df_out, merged = await review_merge(concepts_out, concept_df_out, concept_col, concept_col_prefix, model_name=model_name, sess=sess)
+    concepts_out, concept_df_out, removed = await review_remove(concepts, concept_df, concept_col_prefix, model_name=model_name, sess=sess)
+    concepts_out, concept_df_out, merged = await review_merge(concepts_out, concept_df_out, concept_col_prefix, model_name=model_name, sess=sess)
 
+    merge_results = get_merge_results(merged)
+    logs = f"""
+    Auto-review:
+    Removed ({len(removed)}):
+        {removed}
+    Merged ({len(merged)}): 
+    {merge_results}
+    """
     if debug:
-        print(f"Removed ({len(removed)}):\n{removed}")
-        print(f"Merged ({len(merged)}):")
-        pretty_print_merge_results(merged)
+        print(f"\n\nAuto-review")
+        print(f"")
+        print(f"")
+        
 
     # TODO: ask model to filter to the "best" N concepts
+    if sess is not None:
+        sess.concepts = concepts_out
+    
+    if return_logs:
+        return concepts_out, concept_df_out, logs
     return concepts_out, concept_df_out
 
 
@@ -426,15 +463,14 @@ async def review(concepts, concept_df, concept_col, concept_col_prefix, model_na
 # Input: concept_df (columns: doc_id, text, concept_id, concept_name, concept_prompt)
 # Parameters: n_concepts
 # Output: 
-# - concepts: dict (concept_id -> concept dict)
+# - concepts: dict (concept_id -> Concept)
 # - concept_df: DataFrame (columns: doc_id, text, concept_id, concept_name, concept_prompt)
-async def review_remove(concepts, concept_df, concept_col, concept_col_prefix, model_name, sess):
+async def review_remove(concepts, concept_df, concept_col_prefix, model_name, sess):
     concepts = concepts.copy()  # Make a copy of the concepts dict to avoid modifying the original
     start = time.time()
     concept_name_col = f"{concept_col_prefix}_name"
 
-    concepts_list = concept_df[concept_col].tolist()
-    concepts_list = [f"- {c}" for c in concepts_list]
+    concepts_list = [f"- Name: {c.name}, Prompt: {c.prompt}" for c in concepts.values()]
     concepts_list_str = "\n".join(concepts_list)
     arg_dicts = [{
         "themes": concepts_list_str,
@@ -442,27 +478,34 @@ async def review_remove(concepts, concept_df, concept_col, concept_col_prefix, m
 
     # Run prompts
     prompt_template = review_remove_prompt
-    res_text, res_full = await multi_query_gpt_wrapper(prompt_template, arg_dicts, model_name)
+    if sess is not None:
+        rate_limits = sess.rate_limits
+    else:
+        rate_limits = None
+    res_text, res_full = await multi_query_gpt_wrapper(prompt_template, arg_dicts, model_name, rate_limits=rate_limits)
 
     # Process results
     res = res_text[0]
     concepts_to_remove = json_load(res, top_level_key="remove")
 
-    concept_df_out = concept_df.copy()
-    concept_df_out = concept_df_out[~concept_df_out[concept_name_col].isin(concepts_to_remove)]
+    if concept_df is not None:
+        concept_df_out = concept_df.copy()
+        concept_df_out = concept_df_out[~concept_df_out[concept_name_col].isin(concepts_to_remove)]  # Remove from concept_df
+    else:
+        concept_df_out = None
     c_ids_to_remove = []
     for c_id, c in concepts.items():
-        if c['name'] in concepts_to_remove:
+        if c.name in concepts_to_remove:
             c_ids_to_remove.append(c_id)
     for c_id in c_ids_to_remove:
-        concepts.pop(c_id, None)
+        concepts.pop(c_id, None)  # Remove from concepts dict
 
-    save_progress(sess, concept_df_out, step_name="review_remove", start=start, res=res_full, model_name=model_name)
+    save_progress(sess, concept_df_out, step_name="Review-remove", start=start, res=res_full, model_name=model_name)
     return concepts, concept_df_out, concepts_to_remove
 
 def get_concept_by_name(concepts, concept_name):
     for c_id, c in concepts.items():
-        if c['name'] == concept_name:
+        if c.name == concept_name:
             return c_id, c
     return None, None
 
@@ -470,16 +513,16 @@ def get_concept_by_name(concepts, concept_name):
 # Input: concept_df (columns: doc_id, text, concept_id, concept_name, concept_prompt)
 # Parameters: n_concepts
 # Output: 
-# - concepts: dict (concept_id -> concept dict)
+# - concepts: dict (concept_id -> Concept)
 # - concept_df: DataFrame (columns: doc_id, text, concept_id, concept_name, concept_prompt)
-async def review_merge(concepts, concept_df, concept_col, concept_col_prefix, model_name, sess):
+async def review_merge(concepts, concept_df, concept_col_prefix, model_name, sess):
     concepts = concepts.copy()  # Make a copy of the concepts dict to avoid modifying the original
     start = time.time()
+    concept_col = concept_col_prefix
     concept_name_col = f"{concept_col_prefix}_name"
     concept_prompt_col = f"{concept_col_prefix}_prompt"
 
-    concepts_list = concept_df[concept_col].tolist()
-    concepts_list = [f"- {c}" for c in concepts_list]
+    concepts_list = [f"- Name: {c.name}, Prompt: {c.prompt}" for c in concepts.values()]
     concepts_list_str = "\n".join(concepts_list)
     arg_dicts = [{
         "themes": concepts_list_str,
@@ -487,13 +530,20 @@ async def review_merge(concepts, concept_df, concept_col, concept_col_prefix, mo
 
     # Run prompts
     prompt_template = review_merge_prompt
-    res_text, res_full = await multi_query_gpt_wrapper(prompt_template, arg_dicts, model_name)
+    if sess is not None:
+        rate_limits = sess.rate_limits
+    else:
+        rate_limits = None
+    res_text, res_full = await multi_query_gpt_wrapper(prompt_template, arg_dicts, model_name, rate_limits=rate_limits)
 
     # Process results
     res = res_text[0]
     concepts_to_merge = json_load(res, top_level_key="merge")
 
-    concept_df_out = concept_df.copy()
+    if concept_df is not None:
+        concept_df_out = concept_df.copy()
+    else:
+        concept_df_out = None
 
     # Remove all original concepts
     # Add new merged concepts
@@ -518,30 +568,56 @@ async def review_merge(concepts, concept_df, concept_col, concept_col_prefix, mo
             c_id, c = get_concept_by_name(concepts, orig_concept)
             if c is not None:
                 c_ids_to_remove.append(c_id)
-                merged_example_ids.extend(c["example_ids"])
+                merged_example_ids.extend(c.example_ids)
         
         # Create new merged concept in dict
         new_concept_name = merge_result["merged_theme_name"]
         new_concept_prompt = merge_result["merged_theme_prompt"]
         new_concept_id = str(uuid.uuid4())
-        concepts[new_concept_id] = {
-            "name": new_concept_name,
-            "prompt": new_concept_prompt,
-            "example_ids": merged_example_ids,
-        }
+        concepts[new_concept_id] = Concept(name=new_concept_name, prompt=new_concept_prompt, example_ids=merged_example_ids, active=False)
 
         # Replace prior df row with new merged concept
-        for orig_concept in orig_concepts:
-            concept_df_out.loc[concept_df_out[concept_name_col]==orig_concept, concept_name_col] = new_concept_name
-            concept_df_out.loc[concept_df_out[concept_name_col]==orig_concept, concept_prompt_col] = new_concept_prompt
-            concept_df_out.loc[concept_df_out[concept_name_col]==orig_concept, concept_col] = f"{new_concept_name}: {new_concept_prompt}"
+        if concept_df is not None:
+            for orig_concept in orig_concepts:  # Merge in concept_df
+                concept_df_out.loc[concept_df_out[concept_name_col]==orig_concept, concept_name_col] = new_concept_name
+                concept_df_out.loc[concept_df_out[concept_name_col]==orig_concept, concept_prompt_col] = new_concept_prompt
+                concept_df_out.loc[concept_df_out[concept_name_col]==orig_concept, concept_col] = f"{new_concept_name}: {new_concept_prompt}"
         
     for c_id in c_ids_to_remove:
-        concepts.pop(c_id, None)
+        concepts.pop(c_id, None) # Remove from concepts dict
 
-    save_progress(sess, concept_df_out, step_name="review_merge", start=start, res=res_full, model_name=model_name)
+    save_progress(sess, concept_df_out, step_name="Review-merge", start=start, res=res_full, model_name=model_name)
     return concepts, concept_df_out, concepts_to_merge
 
+# Model selects the best concepts up to `max_concepts`
+# Input: concepts (concept_id -> Concept)
+# Parameters: max_concepts, model_name
+# Output: selected_concepts: dict (concept_id -> Concept)
+async def review_select(concepts, max_concepts, model_name, rate_limits=None):
+    concepts_list = [f"- Name: {c.name}, Prompt: {c.prompt}" for c in concepts.values()]
+    concepts_list_str = "\n".join(concepts_list)
+    arg_dicts = [{
+        "themes": concepts_list_str,
+        "max_concepts": max_concepts,
+    }]
+
+    # Run prompts
+    prompt_template = review_select_prompt
+    res_text, res_full = await multi_query_gpt_wrapper(prompt_template, arg_dicts, model_name, rate_limits=rate_limits)
+
+    # Process results
+    res = res_text[0]
+    selected_concept_names = json_load(res, top_level_key="selected")
+
+    selected_concepts = []
+    for c_id, c in concepts.items():
+        if c.name in selected_concept_names:
+            selected_concepts.append(c_id)
+
+    if len(selected_concepts) > max_concepts:
+        selected_concepts = selected_concepts[:max_concepts]
+    return selected_concepts
+
 
 def get_ex_batch_args(df, text_col, doc_id_col, concept_name, concept_prompt):
     ex = get_examples_dict(df, text_col, doc_id_col)
@@ -635,7 +711,7 @@ def get_empty_score_df(in_df, concept, concept_id, text_col, doc_id_col):
     return out_df[SCORE_DF_OUT_COLS]
 
 # Performs scoring for one concept
-async def score_helper(concept, batch_i, concept_id, df, text_col, doc_id_col, model_name, batch_size, get_highlights, sess):
+async def score_helper(concept, batch_i, concept_id, df, text_col, doc_id_col, model_name, batch_size, get_highlights, sess, threshold):
     # TODO: add support for only a sample of examples
     # TODO: set consistent concept IDs for reference
 
@@ -653,7 +729,11 @@ async def score_helper(concept, batch_i, concept_id, df, text_col, doc_id_col, m
         prompt_template = score_highlight_prompt
     else:
         prompt_template = score_no_highlight_prompt
-    results, res_full = await multi_query_gpt_wrapper(prompt_template, arg_dicts, model_name, batch_num=batch_i)
+    if sess is not None:
+        rate_limits = sess.rate_limits
+    else:
+        rate_limits = None
+    results, res_full = await multi_query_gpt_wrapper(prompt_template, arg_dicts, model_name, batch_num=batch_i, rate_limits=rate_limits)
 
     # Parse results
     # Cols: doc_id, text, concept_id, concept_name, concept_prompt, score, highlight
@@ -673,13 +753,13 @@ async def score_helper(concept, batch_i, concept_id, df, text_col, doc_id_col, m
         missing_rows = get_empty_score_df(missing_rows, concept, concept_id, text_col, doc_id_col)
         score_df = pd.concat([score_df, missing_rows])
 
-    save_progress(sess, score_df, step_name="score_helper", start=None, res=res_full, model_name=model_name)
+    save_progress(sess, score_df, step_name="Score-helper", start=None, res=res_full, model_name=model_name)
     if sess is not None:
         # Save to session if provided
         sess.results[concept_id] = score_df
     
         # Generate summary
-        cur_summary = await summarize_concept(score_df, concept_id, model_name, sess=sess)
+        cur_summary = await summarize_concept(score_df, concept_id, model_name, sess=sess, threshold=threshold)
     return score_df
 
 # Performs scoring for all concepts
@@ -687,7 +767,7 @@ async def score_helper(concept, batch_i, concept_id, df, text_col, doc_id_col, m
 #   --> text could be original, filtered (quotes), and/or summarized (bullets)
 # Parameters: threshold
 # Output: score_df (columns: doc_id, text, concept_id, concept_name, concept_prompt, score, highlight)
-async def score_concepts(text_df, text_col, doc_id_col, concepts, model_name="gpt-3.5-turbo", batch_size=5, get_highlights=False, sess=None):
+async def score_concepts(text_df, text_col, doc_id_col, concepts, model_name="gpt-3.5-turbo", batch_size=5, get_highlights=False, sess=None, threshold=1.0):
     # Scoring operates on "text" column for each concept
     start = time.time()
 
@@ -696,13 +776,13 @@ async def score_concepts(text_df, text_col, doc_id_col, concepts, model_name="gp
     text_df = filter_empty_rows(text_df, text_col)
 
     text_df[doc_id_col] = text_df[doc_id_col].astype(str)
-    tasks = [score_helper(concept, concept_i, concept_id, text_df, text_col, doc_id_col, model_name, batch_size, get_highlights, sess=sess) for concept_i, (concept_id, concept) in enumerate(concepts.items())]
+    tasks = [score_helper(concept, concept_i, concept_id, text_df, text_col, doc_id_col, model_name, batch_size, get_highlights, sess=sess, threshold=threshold) for concept_i, (concept_id, concept) in enumerate(concepts.items())]
     score_dfs = await tqdm_asyncio.gather(*tasks, file=sys.stdout)
 
     # Combine score_dfs
     score_df = pd.concat(score_dfs)
 
-    save_progress(sess, score_df, step_name="score", start=start, res=None, model_name=None)
+    save_progress(sess, score_df, step_name="Score", start=start, res=None, model_name=None)
     return score_df
 
 # Based on concept scoring, refine concepts
@@ -745,7 +825,7 @@ def refine(score_df, concepts, threshold=1, generic_threshold=0.75, rare_thresho
     
     return concepts
 
-async def summarize_concept(score_df, concept_id, model_name="gpt-4-turbo-preview", sess=None, threshold=0.75, summary_length="15-20 word", score_col="score", highlight_col="highlight"):
+async def summarize_concept(score_df, concept_id, model_name="gpt-4-turbo-preview", sess=None, threshold=1.0, summary_length="15-20 word", score_col="score", highlight_col="highlight"):
     # Summarizes behavior in each concept
     df = score_df.copy()
     df = df[df[score_col] >= threshold]
@@ -769,7 +849,11 @@ async def summarize_concept(score_df, concept_id, model_name="gpt-4-turbo-previe
     
     # Run prompts
     prompt_template = summarize_concept_prompt
-    res_text, res_full = await multi_query_gpt_wrapper(prompt_template, arg_dicts, model_name)
+    if sess is not None:
+        rate_limits = sess.rate_limits
+    else:
+        rate_limits = None
+    res_text, res_full = await multi_query_gpt_wrapper(prompt_template, arg_dicts, model_name, rate_limits=rate_limits)
 
     # Process results
     res = res_text[0]
@@ -782,7 +866,7 @@ async def summarize_concept(score_df, concept_id, model_name="gpt-4-turbo-previe
     return None
 
 # Returns ids of not-covered documents
-def get_not_covered(score_df, doc_id_col, threshold=0.75, debug=False):
+def get_not_covered(score_df, doc_id_col, threshold=1.0, debug=False):
     # Convert to thresholded scores; sum scores across concepts for each doc
     df = score_df.copy()
     df["score"] = df["score"].apply(lambda x: 1 if x >= threshold else 0)
@@ -798,7 +882,7 @@ def get_not_covered(score_df, doc_id_col, threshold=0.75, debug=False):
     return doc_ids
 
 # Returns ids of covered-by-generic documents
-def get_covered_by_generic(score_df, doc_id_col, threshold=0.75, generic_threshold=0.5, debug=False):
+def get_covered_by_generic(score_df, doc_id_col, threshold=1.0, generic_threshold=0.5, debug=False):
     # Determines generic concepts
     df = score_df.copy()
     df["score"] = df["score"].apply(lambda x: 1 if x >= threshold else 0)
@@ -894,7 +978,11 @@ async def auto_eval(items, concepts, model_name="gpt-3.5-turbo", debug=False, se
 
     # Run prompts
     prompt_template = concept_auto_eval_prompt
-    res_text, res_full = await multi_query_gpt_wrapper(prompt_template, arg_dicts, model_name)
+    if sess is not None:
+        rate_limits = sess.rate_limits
+    else:
+        rate_limits = None
+    res_text, res_full = await multi_query_gpt_wrapper(prompt_template, arg_dicts, model_name, rate_limits=rate_limits)
     
     res_text = res_text[0]
     if debug:
@@ -943,7 +1031,7 @@ async def auto_eval(items, concepts, model_name="gpt-3.5-turbo", debug=False, se
         else:
             print(f"{c_id}: {c} -- NONE")
 
-    save_progress(sess, df=None, step_name="auto_eval", start=start, res=res_full, model_name=model_name)
+    save_progress(sess, df=None, step_name="Auto-eval", start=start, res=res_full, model_name=model_name)
     return concepts_found, concept_coverage
 
 # Adds color-background styling for score columns to match score value
@@ -962,8 +1050,24 @@ def format_bullets(orig, add_quotes=False):
         lines = [f"<li>{line}</li>" for line in orig]
     return "<ul>" + "".join(lines) + "</ul>"
 
+def remove_emojis(string):
+    emoji_pattern = re.compile(
+        "["
+        u"\U0001F600-\U0001F64F" # emoticons
+        u"\U0001F300-\U0001F5FF" # symbols & pictographs
+        u"\U0001F680-\U0001F6FF" # transport & map symbols
+        u"\U0001F1E0-\U0001F1FF" # flags (iOS)
+        u"\U00002702-\U000027B0"
+        u"\U000024C2-\U0001F251"
+        "]+", 
+        flags=re.UNICODE
+    )
+    
+    return emoji_pattern.sub(r'', string)
+
 # Adds color-background styling for highlight columns to match score value
 def format_highlight(orig: str, quotes, x: float):
+        orig = remove_emojis(orig)
         color_str = f"rgba(130, 193, 251, {x*0.5})"
         start_tag = f"<span style='background-color: {color_str}'>"
         end_tag = "</span>"
@@ -1071,7 +1175,7 @@ def get_groupings(df, slice_col, max_slice_bins, slice_bounds):
 # Helper function for `visualize()` to generate the underlying dataframes
 # Parameters:
 # - threshold: float (minimum score of positive class)
-def prep_vis_dfs(df, score_df, doc_id_col, doc_col, score_col, df_filtered, df_bullets, concepts, cols_to_show, slice_col, max_slice_bins, slice_bounds, show_highlights, norm_by=None, debug=False, threshold=None, outlier_threshold=0.75):
+def prep_vis_dfs(df, score_df, doc_id_col, doc_col, score_col, df_filtered, df_bullets, concepts, cols_to_show, slice_col, max_slice_bins, slice_bounds, show_highlights, norm_by=None, debug=False, threshold=None, outlier_threshold=1.0):
     # TODO: codebook info
 
     # Handle groupings
@@ -1110,7 +1214,7 @@ def prep_vis_dfs(df, score_df, doc_id_col, doc_col, score_col, df_filtered, df_b
     rationale_col = "score rationale"
     highlight_col = "highlight"
     rationale_df = score_df[[doc_id_col, "concept_name", "rationale", highlight_col]]
-    rationale_df.rename(columns={"rationale": rationale_col}, inplace=True)
+    rationale_df = rationale_df.rename(columns={"rationale": rationale_col})
     rationale_df[doc_id_col] = rationale_df[doc_id_col].astype(str)
 
     # Prep data for each group
@@ -1198,7 +1302,7 @@ def prep_vis_dfs(df, score_df, doc_id_col, doc_col, score_col, df_filtered, df_b
         for concept in concept_names:
             cur_scores = [clean_score(x, threshold) for x in cur_df[concept].tolist()]
             if len(cur_scores) > 0:
-                matches = [x for x in cur_scores if x > outlier_threshold]
+                matches = [x for x in cur_scores if x >= outlier_threshold]
                 n_matches = len(matches)
             else:
                 n_matches = 0
