@@ -4,11 +4,12 @@
 # Imports
 import time
 import pandas as pd
-import ipywidgets as widgets
 import random
 from nltk.tokenize import sent_tokenize
 import os
-import openai
+from yaspin import yaspin
+import base64
+import requests
 
 
 # Local imports
@@ -16,12 +17,12 @@ if __package__ is None or __package__ == '':
     # uses current directory visibility
     from concept_induction import *
     from concept import Concept
-    from llm import get_token_estimate, EMBED_COSTS
+    from llm import get_token_estimate, EMBED_COSTS, RATE_LIMITS
 else:
     # uses current package visibility
     from .concept_induction import *
     from .concept import Concept
-    from .llm import get_token_estimate, EMBED_COSTS
+    from .llm import get_token_estimate, EMBED_COSTS, RATE_LIMITS
 
 # WORKBENCH class ================================
 class lloom:
@@ -30,17 +31,33 @@ class lloom:
         df: pd.DataFrame,
         text_col: str,
         id_col: str = None,
-        save_path: str = None,
+        distill_model_name = "gpt-3.5-turbo",
+        embed_model_name = "text-embedding-3-large",
+        synth_model_name = "gpt-4-turbo",
+        score_model_name = "gpt-3.5-turbo",
+        rate_limits = {}, # D_i = "model-name": (n_requests, wait_time_secs)
         debug: bool = False,
     ):
         # Settings
-        self.model_name = "gpt-3.5-turbo"
-        self.embed_model_name = "text-embedding-3-large"
-        self.synth_model_name = "gpt-4-turbo-preview"
-        self.score_model_name = "gpt-3.5-turbo"
-        self.use_base_api = True
+        self.distill_model_name = distill_model_name  # Distill operators (filter and summarize)
+        self.embed_model_name = embed_model_name  # Cluster operator
+        self.synth_model_name = synth_model_name  # Synthesize operator
+        self.score_model_name = score_model_name  # Score operator
         self.debug = debug  # Whether to run in debug mode
 
+        # Rate limits
+        # n_requests: number of requests allowed in one batch
+        # wait_time_secs: time period (in seconds) to wait before making more requests
+        # RPM (Requests per minute) = n_requests * (60 / wait_time_secs)
+        if len(rate_limits) == 0:
+            rate_limits = RATE_LIMITS  # Use default values set from llm.py
+        else:
+            # Intersect user-provided rate_limits with full set of options set in default RATE_LIMITS
+            for k, v in RATE_LIMITS.items():
+                if k not in rate_limits:
+                    rate_limits[k] = v  # Add in defaults for any missing values
+        self.rate_limits = rate_limits
+
         # Input data
         self.doc_id_col = id_col
         self.doc_col = text_col
@@ -65,11 +82,9 @@ class lloom:
             "out_tokens": [],
         }
 
-        # Set up API key
+        # Check for API key
         if "OPENAI_API_KEY" not in os.environ:
             raise Exception("API key not found. Please set the OPENAI_API_KEY environment variable by running: `os.environ['OPENAI_API_KEY'] = 'your_key'`")
-        else:
-            openai.api_key = os.environ["OPENAI_API_KEY"]
     
     # Preprocesses input dataframe
     def preprocess_df(self, df):
@@ -104,12 +119,46 @@ class lloom:
         print(f"Saved session to {cur_path}")
 
         self.select_widget = select_widget  # Restore widget after saving
+    
+    def get_pkl_str(self):
+        # Saves current session to pickle string
+        select_widget = self.select_widget
+        self.select_widget = None  # Remove widget before saving (can't be pickled)
+
+        pkl_str = pickle.dumps(self)
+        self.select_widget = select_widget  # Restore widget after saving
+        return pkl_str
 
     def get_save_key(self, step_name):
         t = time.strftime('%Y-%m-%d-%H-%M-%S', time.localtime())
         k = (step_name, t)  # Key of step name and current time
         return k
 
+    # Printed text formatting
+    def bold_txt(self, s):
+        # Bold text
+        return f"\033[1m{s}\033[0m"
+
+    def highlight_txt(self, s, color="yellow"):
+        # Highlight text (background color)
+        if color == "yellow":
+            return f"\x1b[48;5;228m{s}\x1b[0m"
+        elif color == "blue":
+            return f"\x1b[48;5;117m{s}\x1b[0m"
+
+    def bold_highlight_txt(self, s):
+        # Both bold and highlight text
+        return self.bold_txt(self.highlight_txt(s))
+    
+    def print_step_name(self, step_name):
+        # Print step name (with blue highlighting)
+        format_step_name = f"{self.highlight_txt(step_name, color='blue')}"
+        print(f"\n\n{format_step_name}")
+
+    def spinner_wrapper(self):
+        # Wrapper for loading spinner
+        return yaspin(text="Loading")
+
     # Estimate cost of generation for the given params
     def estimate_gen_cost(self, params=None, verbose=False):
         if params is None:
@@ -125,17 +174,17 @@ class lloom:
         # Filter: generate filter_n_quotes for each doc
         est_cost = {}
         
-        filter_in_tokens = np.sum([get_token_estimate(filter_prompt + doc, self.model_name) for doc in self.in_df[self.doc_col].tolist()])
+        filter_in_tokens = np.sum([get_token_estimate(filter_prompt + doc, self.distill_model_name) for doc in self.in_df[self.doc_col].tolist()])
         quotes_tokens_per_doc = params["filter_n_quotes"] * est_quote_tokens 
         filter_out_tokens = quotes_tokens_per_doc * len(self.in_df)
-        est_cost["distill_filter"] = calc_cost_by_tokens(self.model_name, filter_in_tokens, filter_out_tokens)
+        est_cost["distill_filter"] = calc_cost_by_tokens(self.distill_model_name, filter_in_tokens, filter_out_tokens)
 
         # Summarize: create n_bullets for each doc
-        summ_prompt_tokens = get_token_estimate(summarize_prompt, self.model_name)
+        summ_prompt_tokens = get_token_estimate(summarize_prompt, self.distill_model_name)
         summ_in_tokens = np.sum([(summ_prompt_tokens + quotes_tokens_per_doc) for _ in range(len(self.in_df))])
         bullets_tokens_per_doc = params["summ_n_bullets"] * est_bullet_tokens
         summ_out_tokens = bullets_tokens_per_doc * len(self.in_df)
-        est_cost["distill_summarize"] = calc_cost_by_tokens(self.model_name, summ_in_tokens, summ_out_tokens)
+        est_cost["distill_summarize"] = calc_cost_by_tokens(self.distill_model_name, summ_in_tokens, summ_out_tokens)
 
         # Cluster: embed each bullet point
         cluster_rate = EMBED_COSTS[self.embed_model_name] 
@@ -148,9 +197,14 @@ class lloom:
         synth_in_tokens = np.sum([(synth_prompt_tokens + (est_bullet_tokens * n_bullets_per_cluster)) for _ in range(est_n_clusters)])
         synth_out_tokens = params["synth_n_concepts"] * est_n_clusters * est_concept_tokens
         est_cost["synthesize"] = calc_cost_by_tokens(self.synth_model_name, synth_in_tokens, synth_out_tokens)
+
+        # Review: pass all names and prompts
+        rev_in_tokens = synth_out_tokens * 2  # For both review_remove and review_merge
+        rev_out_tokens = rev_in_tokens * 0.5  # Conservatively assume half size
+        est_cost["review"] = calc_cost_by_tokens(self.synth_model_name, rev_in_tokens, rev_out_tokens)
         
         total_cost = np.sum([c[0] + c[1] for c in est_cost.values()])
-        print(f"Estimated cost: {np.round(total_cost, 2)}")
+        print(f"\n\n{self.bold_txt('Estimated cost')}: ${np.round(total_cost, 2)}")
         print("**Please note that this is only an approximate cost estimate**")
 
         if verbose:
@@ -181,8 +235,8 @@ class lloom:
         est_cost = calc_cost_by_tokens(self.score_model_name, score_in_tokens, score_out_tokens)
 
         total_cost = np.sum(est_cost)
-        print(f"Scoring {n_concepts} concepts for {len(self.in_df)} documents")
-        print(f"Estimated cost: {np.round(total_cost, 2)}")
+        print(f"\n\nScoring {n_concepts} concepts for {len(self.in_df)} documents")
+        print(f"{self.bold_txt('Estimated cost')}: ${np.round(total_cost, 2)}")
         print("**Please note that this is only an approximate cost estimate**")
 
         if verbose:
@@ -218,45 +272,46 @@ class lloom:
         }
         return params
     
-    def summary(self, show_detail=True):
+    def summary(self, verbose=True):
         # Time
         total_time = np.sum(list(self.time.values()))
-        print(f"Total time: {total_time:0.2f} sec ({(total_time/60):0.2f} min)")
-        if show_detail:
+        print(f"{self.bold_txt('Total time')}: {total_time:0.2f} sec ({(total_time/60):0.2f} min)")
+        if verbose:
             for step_name, time in self.time.items():
                 print(f"\t{step_name}: {time:0.2f} sec")
 
         # Cost
         total_cost = np.sum(list(self.cost.values()))
-        print(f"\n\nTotal cost: {total_cost:0.2f}")
-        if show_detail:
+        print(f"\n\n{self.bold_txt('Total cost')}: ${total_cost:0.2f}")
+        if verbose:
             for step_name, cost in self.cost.items():
-                print(f"\t{step_name}: {cost:0.2f}")
+                print(f"\t{step_name}: ${cost:0.3f}")
 
         # Tokens
         in_tokens = np.sum(self.tokens["in_tokens"])
         out_tokens = np.sum(self.tokens["out_tokens"])
         total_tokens =  in_tokens + out_tokens
-        print(f"\n\nTokens: total={total_tokens}, in={in_tokens}, out={out_tokens}")
+        print(f"\n\n{self.bold_txt('Tokens')}: total={total_tokens}, in={in_tokens}, out={out_tokens}")
 
     def show_selected(self):
         active_concepts = self.__get_active_concepts()
-        print(f"Active concepts (n={len(active_concepts)}):")
+        print(f"\n\n{self.bold_txt('Active concepts')} (n={len(active_concepts)}):")
         for c_id, c in active_concepts.items():
-            print(f"- {c.name}: {c.prompt}")
+            print(f"- {self.bold_txt(c.name)}: {c.prompt}")
 
     # HELPER FUNCTIONS ================================
-    async def gen(self, seed=None, params=None, n_synth=1, debug=True):
+    async def gen(self, seed=None, params=None, n_synth=1, auto_review=True, debug=True):
         if params is None:
             params = self.auto_suggest_parameters(debug=debug)
             if debug:
-                print(f"Auto-suggested parameters: {params}")
+                print(f"{self.bold_txt('Auto-suggested parameters')}: {params}")
         
         # Run cost estimation
         self.estimate_gen_cost(params)
         
         # Confirm to proceed
-        user_input = input("\nProceed with generation? (y/n): ")
+        print(f"\n\n{self.bold_highlight_txt('Action required')}")
+        user_input = input("Proceed with generation? (y/n): ")
         if user_input.lower() != "y":
             print("Cancelled generation")
             return
@@ -264,36 +319,44 @@ class lloom:
         # Run concept generation
         filter_n_quotes = params["filter_n_quotes"]
         if filter_n_quotes > 1:
-            df_filtered = await distill_filter(
-                text_df=self.in_df, 
-                doc_col=self.doc_col,
-                doc_id_col=self.doc_id_col,
-                model_name=self.model_name,
-                n_quotes=params["filter_n_quotes"],
-                seed=seed,
-                sess=self,
-            )
-            self.df_to_score = df_filtered  # Change to use filtered df for concept scoring
-            self.df_filtered = df_filtered
+            step_name = "Distill-filter"
+            self.print_step_name(step_name)
+            with self.spinner_wrapper() as spinner:
+                df_filtered = await distill_filter(
+                    text_df=self.in_df, 
+                    doc_col=self.doc_col,
+                    doc_id_col=self.doc_id_col,
+                    model_name=self.distill_model_name,
+                    n_quotes=params["filter_n_quotes"],
+                    seed=seed,
+                    sess=self,
+                )
+                self.df_to_score = df_filtered  # Change to use filtered df for concept scoring
+                self.df_filtered = df_filtered
+                spinner.text = "Done"
+                spinner.ok("✅")
             if debug:
-                print("df_filtered")
                 display(df_filtered)
         else:
             # Just use original df to generate bullets
             self.df_filtered = self.in_df
         
-        df_bullets = await distill_summarize(
-            text_df=self.df_filtered, 
-            doc_col=self.doc_col,
-            doc_id_col=self.doc_id_col,
-            model_name=self.model_name,
-            n_bullets=params["summ_n_bullets"],
-            seed=seed,
-            sess=self,
-        )
-        self.df_bullets = df_bullets
+        step_name = "Distill-summarize"
+        self.print_step_name(step_name)
+        with self.spinner_wrapper() as spinner:
+            df_bullets = await distill_summarize(
+                text_df=self.df_filtered, 
+                doc_col=self.doc_col,
+                doc_id_col=self.doc_id_col,
+                model_name=self.distill_model_name,
+                n_bullets=params["summ_n_bullets"],
+                seed=seed,
+                sess=self,
+            )
+            self.df_bullets = df_bullets
+            spinner.text = "Done"
+            spinner.ok("✅")
         if debug:
-            print("df_bullets")
             display(df_bullets)
         
         df_cluster_in = df_bullets
@@ -303,41 +366,73 @@ class lloom:
         # Perform synthesize step n_synth times
         for i in range(n_synth):
             self.concepts = {}
-            df_cluster = await cluster(
-                text_df=df_cluster_in, 
-                doc_col=synth_doc_col,
-                doc_id_col=self.doc_id_col,
-                embed_model_name=self.embed_model_name,
-                sess=self,
-            )
+
+            step_name = "Cluster"
+            self.print_step_name(step_name)
+            with self.spinner_wrapper() as spinner:
+                df_cluster = await cluster(
+                    text_df=df_cluster_in, 
+                    doc_col=synth_doc_col,
+                    doc_id_col=self.doc_id_col,
+                    embed_model_name=self.embed_model_name,
+                    sess=self,
+                )
+                spinner.text = "Done"
+                spinner.ok("✅")
             if debug:
-                print("df_cluster")
                 display(df_cluster)
             
-            df_concepts = await synthesize(
-                cluster_df=df_cluster, 
-                doc_col=synth_doc_col,
-                doc_id_col=self.doc_id_col,
-                model_name=self.synth_model_name,
-                concept_col_prefix=concept_col_prefix,
-                n_concepts=synth_n_concepts,
-                pattern_phrase="unique topic",
-                seed=seed,
-                sess=self,
-            )
-
-            self.concept_history[i] = self.concepts
+            step_name = "Synthesize"
+            self.print_step_name(step_name)
+            with self.spinner_wrapper() as spinner:
+                df_concepts, synth_logs = await synthesize(
+                    cluster_df=df_cluster, 
+                    doc_col=synth_doc_col,
+                    doc_id_col=self.doc_id_col,
+                    model_name=self.synth_model_name,
+                    concept_col_prefix=concept_col_prefix,
+                    n_concepts=synth_n_concepts,
+                    pattern_phrase="unique topic",
+                    seed=seed,
+                    sess=self,
+                    return_logs=True,
+                )
+                spinner.text = "Done"
+                spinner.ok("✅")
             if debug:
-                # Print results
-                print(f"synthesize {i + 1}: (n={len(self.concepts)})")
-                for k, c in self.concepts.items():
-                    print(f'- Concept {k}:\n\t{c.name}\n\t- Prompt: {c.prompt}')
+                print(synth_logs)
+
+                # Review current concepts (remove low-quality, merge similar)
+                if auto_review:
+                    step_name = "Review"
+                    self.print_step_name(step_name)
+                    with self.spinner_wrapper() as spinner:
+                        _, df_concepts, review_logs = await review(
+                            concepts=self.concepts, 
+                            concept_df=df_concepts, 
+                            concept_col_prefix=concept_col_prefix, 
+                            model_name=self.synth_model_name, 
+                            sess=self,
+                            return_logs=True,
+                        )
+                        spinner.text = "Done"
+                        spinner.ok("✅")
+                    if debug:
+                        print(review_logs)
+
+                self.concept_history[i] = self.concepts
+                if debug:
+                    # Print results
+                    print(f"\n\n{self.highlight_txt('Synthesize', color='blue')} {i + 1}: (n={len(self.concepts)} concepts)")
+                    for k, c in self.concepts.items():
+                        print(f'- Concept {k}:\n\t{c.name}\n\t- Prompt: {c.prompt}')
             
             # Update synthesize params for next iteration
             df_concepts["synth_doc_col"] = df_concepts[f"{concept_col_prefix}_name"] + ": " + df_concepts[f"{concept_col_prefix}_prompt"]
             df_cluster_in = df_concepts
             synth_doc_col = "synth_doc_col"
             synth_n_concepts = math.floor(synth_n_concepts * 0.75)
+        print("✅ Done with concept generation!")
 
     def __concepts_to_json(self):
         concept_dict = {c_id: c.to_dict() for c_id, c in self.concepts.items()}
@@ -356,6 +451,20 @@ class lloom:
         self.select_widget = w
         return w
 
+    async def select_auto(self, max_concepts):
+        # Select the best concepts up to max_concepts
+        selected_concepts = await review_select(self.concepts, max_concepts, self.synth_model_name, self.rate_limits)
+
+        # Handle if selection failed
+        if len(selected_concepts) == 0:
+            concept_ids = list(self.concepts.keys())
+            selected_concepts = random.sample(concept_ids, max_concepts)
+
+        # Activate only the selected concepts
+        for c_id in selected_concepts:
+            if c_id in self.concepts:
+                self.concepts[c_id].active = True
+
     def __get_active_concepts(self):
         # Update based on widget
         if self.select_widget is not None:
@@ -388,7 +497,8 @@ class lloom:
         self.estimate_score_cost(n_concepts=len(concepts), batch_size=batch_size, get_highlights=get_highlights)
 
         # Confirm to proceed
-        user_input = input("\nProceed with scoring? (y/n): ")
+        print(f"\n\n{self.bold_highlight_txt('Action required')}")
+        user_input = input("Proceed with scoring? (y/n): ")
         if user_input.lower() != "y":
             print("Cancelled scoring")
             return
@@ -403,8 +513,10 @@ class lloom:
             batch_size=batch_size,
             get_highlights=get_highlights,
             sess=self,
+            threshold=1.0,
         )
 
+        print("✅ Done with concept scoring!")
         return score_df
 
     def __get_concept_from_name(self, name):
@@ -422,13 +534,13 @@ class lloom:
         score_df = score_df.rename(columns={"doc_id": self.doc_id_col})
         return score_df
 
-    def __get_concept_highlights(self, c, threshold=0.75, highlight_col="highlight", lim=3):
+    def __get_concept_highlights(self, c, threshold=1.0, highlight_col="highlight", lim=3):
         if c.name == "Outlier":
             return []
         if c.id not in self.results:
             return []
         score_df = self.results[c.id].copy()
-        score_df = score_df[score_df["score"] > threshold]
+        score_df = score_df[score_df["score"] >= threshold]
         highlights = score_df[highlight_col].tolist()
         # shuffle highlights
         random.shuffle(highlights)
@@ -447,10 +559,10 @@ class lloom:
         ex = df[df[self.doc_id_col].isin(ex_ids)][self.doc_col].tolist()
         return ex
 
-    def __get_df_for_export(self, item_df, threshold=0.75, include_outliers=False):
+    def __get_df_for_export(self, item_df, threshold=1.0, include_outliers=False):
         # Prepares a dataframe meant for exporting the current session results
         # Includes concept, criteria, summary, representative examples, prevalence, and highlights
-        matched = item_df[(item_df.concept_score_orig > threshold)]
+        matched = item_df[(item_df.concept_score_orig >= threshold)]
         if not include_outliers:
             matched = matched[item_df.concept != "Outlier"]
 
@@ -474,7 +586,7 @@ class lloom:
     # - show_highlights: bool (whether to show text highlights)
     # - norm_by: str (how to normalize scores: "concept" or "slice")
     # - export_df: bool (whether to return a dataframe for export)
-    def vis(self, cols_to_show=[], slice_col=None, max_slice_bins=5, slice_bounds=None, show_highlights=True, norm_by=None, export_df=False):
+    def vis(self, cols_to_show=[], slice_col=None, max_slice_bins=5, slice_bounds=None, show_highlights=True, norm_by=None, export_df=False, include_outliers=False):
         active_concepts = self.__get_active_concepts()
         score_df = self.get_score_df()
 
@@ -495,18 +607,18 @@ class lloom:
             norm_by=norm_by,
         )
         if export_df:
-            return self.__get_df_for_export(item_df)
+            return self.__get_df_for_export(item_df, include_outliers=include_outliers)
         
         return widget
 
-    def export_df(self):
-        return self.vis(export_df=True)
+    def export_df(self, include_outliers=False):
+        return self.vis(export_df=True, include_outliers=include_outliers)
     
-    def export_json(self, threshold=0.75):
+    def export_json(self, threshold=1.0):
         def format_dict(c):
             c["criteria"] = c["prompt"]
             cur_score_df = self.results[c["id"]]
-            matched = cur_score_df[cur_score_df.score > threshold]
+            matched = cur_score_df[cur_score_df.score >= threshold]
             c["n"] = len(matched)
             # Remove unused columns
             del c["prompt"]
@@ -519,6 +631,59 @@ class lloom:
         concepts_json = json.dumps(concepts_dict)
         return concepts_json
 
+    def submit(self):
+        # Submit the current session results to the database
+        # Prepare pickled lloom instance
+        l_pkl = self.get_pkl_str()
+        l_pkl_str = base64.b64encode(l_pkl).decode('ascii')
+
+        # Gather user inputs
+        print("Thank you for using LLooM and submitting your work! We would love to hear more about your analysis.")
+        print("\nPlease provide a contact email address. This will allow us to follow up to better meet your needs and/or feature your work on our site!")
+        email = input("Email address: ")
+
+        print("\nBriefly summarize the goal of your analysis: What data were you using? What questions were you trying to answer?")
+        goal = input("Goal: ")
+
+        with self.spinner_wrapper() as spinner:
+            cur_data = {
+                "email": email,
+                "goal": goal,
+                "lloom_pkl": l_pkl_str,
+            }
+            hdr = {"Content-Type": "application/json"}
+            cur_url = "https://lloom-log-server.vercel.app"
+
+            try:
+                r = requests.post(f"{cur_url}/save", json=cur_data, headers=hdr)
+
+                # Parse result
+                r_json = json.loads(r.text)
+                status = r_json["message"]
+                spinner.text = f"Submission status: {status}"
+                spinner.ok("✅")
+            except Exception as e:
+                spinner.fail("❌")
+                print(f"Error: {e}")
+
+    async def gen_auto(
+        self,
+        max_concepts=8,
+        seed=None, params=None, n_synth=1,
+        debug=True
+    ):
+        # Runs gen(), select(), and score() all at once
+        # Run generation
+        await self.gen(seed=seed, params=params, n_synth=n_synth, auto_review=True, debug=debug)
+
+        # Select the best concepts
+        await self.select_auto(max_concepts=max_concepts)
+        self.show_selected()
+
+        # Run scoring
+        score_df = await self.score()
+        return score_df
+
     async def add(self, name, prompt, ex_ids=[], get_highlights=True):
         # Add concept
         c = Concept(name=name, prompt=prompt, example_ids=ex_ids, active=True)
